/*
 * Copyright (c) 2021-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * Wrappers around hamlib for ease of use
 */

#include <hamlib/rig.h>

#include <jeffpc/types.h>
#include <jeffpc/error.h>
#include <jeffpc/mem.h>
#include <jeffpc/synch.h>

#include "hamlib.h"

static LOCK_CLASS(lc_rig);

struct hrig {
	struct lock lock;
	RIG *rig;

	bool open;

	struct {
		bool no_txpower;
	} quirks;
};

static int hamlib_err2errno(int e)
{
#define NOT_YET(e) \
	case e: \
		panic("Hamlib error code %s not yet converted by %s", \
		      #e, __func__)

	switch ((enum rig_errcode_e) -e) {
		case RIG_OK:
			return 0;
		case RIG_EINVAL:
			return -EINVAL;
		NOT_YET(RIG_ECONF);
		case RIG_ENOMEM:
			return -ENOMEM;
		case RIG_ENIMPL:
			return -ENOTSUP;
		case RIG_ETIMEOUT:
			return -ETIMEDOUT;
		case RIG_EIO:
			return -EIO;
		NOT_YET(RIG_EINTERNAL);
		case RIG_EPROTO:
			return -EPROTO;
		NOT_YET(RIG_ERJCTED);
		NOT_YET(RIG_ETRUNC);
		NOT_YET(RIG_ENAVAIL);
		NOT_YET(RIG_ENTARGET);
		NOT_YET(RIG_BUSERROR);
		NOT_YET(RIG_BUSBUSY);
		NOT_YET(RIG_EARG);
		NOT_YET(RIG_EVFO);
		case RIG_EDOM:
			return -EDOM;
	}
#undef NOT_YET

	panic("Unknown hamlib error code: %u", e);
}

struct hrig *hamlib_alloc(uint32_t model, const char *dev)
{
	struct hrig *rig;
	int ret;

	/* avoid spamming the console with pointless messages */
	rig_set_debug_level(RIG_DEBUG_NONE);

	rig = malloc(sizeof(struct hrig));
	if (!rig)
		return ERR_PTR(-ENOMEM);

	MXINIT(&rig->lock, &lc_rig);
	rig->open = false;

	rig->rig = rig_init(model);
	if (!rig->rig) {
		cmn_err(CE_ERROR, "Failed to initialize rig");
		ret = -ENOMEM;
		goto err_rig;
	}

	/* TODO: is there a proper way to set the path? */
	strncpy(rig->rig->state.rigport.pathname, dev,
		sizeof(rig->rig->state.rigport.pathname) - 1);

	if (model == RIG_MODEL_NETRIGCTL) {
		rig->quirks.no_txpower = true;
		cmn_err(CE_WARN, "Hamlib 4.3.1 and older has buggy netrigctl");
		cmn_err(CE_WARN, "Disabling power announcements");
	}

	return rig;

err_rig:
	free(rig);

	return ERR_PTR(ret);
}

void hamlib_free(struct hrig *rig)
{
	rig_close(rig->rig);

	MXDESTROY(&rig->lock);

	free(rig);
}

int hamlib_open(struct hrig *rig)
{
	int ret;

	if (rig->open)
		return -EBUSY;

	ret = hamlib_err2errno(rig_open(rig->rig));

	rig->open = (ret == 0);

	return ret;
}

static int hamlib_get_state_unlocked(struct hrig *rig, struct hrig_state *state)
{
	rmode_t tx_mode, rx_mode;
	freq_t tx_freq, rx_freq;
	unsigned int power;
	pbwidth_t width;
	split_t split;
	vfo_t tx_vfo;
	int ret;

	/*
	 *  1. get tx freq & mode
	 *     split=0: \get_freq \get_mode
	 *     split=1: \get_freq \get_mode
	 *  2. get rx freq & mode
	 *     split=0: \get_freq \get_mode
	 *     split=1: \get_split_freq \get_split_mode
	 *  3. get tx power
	 *     \get_level RFPOWER
	 *     \power2mW <0.25> <freq> <mode>
	 */

	/* get rx freq, mode, and split */
	ret = rig_get_freq(rig->rig, RIG_VFO_CURR, &rx_freq);
	if (ret)
		goto err;

	ret = rig_get_mode(rig->rig, RIG_VFO_CURR, &rx_mode, &width);
	if (ret)
		goto err;

	ret = rig_get_split_vfo(rig->rig, RIG_VFO_CURR, &split, &tx_vfo);
	if (ret)
		goto err;

	/* get tx freq and mode */
	if (split == RIG_SPLIT_ON) {
		ret = rig_get_split_freq_mode(rig->rig, RIG_VFO_CURR,
					      &tx_freq, &tx_mode,
					      &width);
		if (ret)
			goto err;
	} else {
		tx_freq = rx_freq;
		tx_mode = rx_mode;
	}

	/*
	 * get tx power
	 *
	 * NOTE: Older versions of hamlib have a buggy netrigctl that may
	 * segfault due to a buffer overflow when calling rig_power2mW, so
	 * we don't risk it and for now send a zero which means "no power
	 * information available".
	 */
	if (rig->quirks.no_txpower) {
		power = 0;
	} else {
		value_t val;

		ret = rig_get_level(rig->rig, RIG_VFO_CURR, RIG_LEVEL_RFPOWER,
				    &val);
		if (ret)
			goto err;

		ret = rig_power2mW(rig->rig, &power, val.f, tx_freq, tx_mode);
		if (ret)
			goto err;
	}

	state->tx_freq = tx_freq;
	state->rx_freq = rx_freq;
	state->power = power;

	return 0;

err:
	return hamlib_err2errno(ret);
}

int hamlib_get_state(struct hrig *rig, struct hrig_state *state)
{
	int ret;

	MXLOCK(&rig->lock);
	ret = hamlib_get_state_unlocked(rig, state);
	MXUNLOCK(&rig->lock);

	return ret;
}
